/*
 * Copyright (c) 2002
 * Ningning Hu and the Carnegie Mellon University.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <sys/param.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>

#include "capture.h"
#include "pcap.h"
#include "gmt2local.h"
#include "inet.h"
#include "setsignal.h"
#include "common.h"

#define PhaseWait	(1)	/* wait for next packets for 1s */
#define NoDataWait	(5)	/* wait time when there is no data coming */
#define MaxClientNum 	(64)	/* this is used for filter cmd set up */
#define MaxNoDataNum	(3)	/* number of phases allowed to be no data */

char version[] = "1.0";

struct filter_item {
	char src_ip_str[16];
	char dst_ip_str[16];
	uint32 src_ip;
	uint32 dst_ip;
	int control_sock;

	struct pkt_rcd_t record[MaxProbeNum];
	/* last active time */
	double pre_time;
	/* the number of packets we have filtered out for this filter */
	int count;
	/* the number of packets client should send to us */
	int probe_num;

	/* number of consecutive no data phase, 3 means dead */
	int nodata_count;
	/* working state */
	int working;

	struct filter_item * next;
};

struct device_item {
	char dev_str[32];
	/* assume one interface only works for one ip address */
	uint32 dst_ip;

	pcap_t * pc;
	pthread_mutex_t pc_mutex;

	struct filter_item * filter_list;
	pthread_mutex_t filter_mutex;

	pthread_t thread;

	struct device_item * next;
};
/* record the device interface which is working on packet filtering */
struct device_item * device_list = NULL;

int sock;
struct sockaddr_in dst_addr, src_addr;
uint32 src_ip, dst_ip;
uint16 src_port, dst_port;
char src_ip_str[16], dst_ip_str[16];
int src_size, dst_size;

int probe_num = 60;
int verbose = 0;
int debug = 0;


/* usage message */
void Usage()
{
	printf("igi-%s server usage:\n\n", version);
	printf("\tigi_server [-vd]\n\n");
	printf("\t-h\tprint this message.\n");
	printf("\t-v\tverbose mode.\n");
	printf("\t-d\tdebug mode (dump more message than the verbose mode).\n\n");
	exit(1);
}

/* make a clean exit on interrupts */
RETSIGTYPE cleanup(int signo)
{
	struct device_item * pdev = device_list;
	struct device_item * pdev_pre = NULL;
	struct filter_item * pfilter;
	struct filter_item * pfilter_post;


	/* client the device list */
	while (pdev != NULL) {
	    pfilter = pdev->filter_list;
	    while (pfilter != NULL) {
		pfilter_post = pfilter->next;
		free(pfilter);
		pfilter = pfilter_post;
	    }

	    pdev_pre = pdev->next;
      	    CaptureEnd(pdev->pc);
	    pdev = pdev_pre;
	}

  	exit(0);
}


/* append a device_item on to device_list */
void append_device(struct device_item * pdev)
{
	struct device_item * pre, *p;

	p = device_list;
	pre = NULL;

	if (p == NULL) {
	    device_list = pdev;
	    return;
	}

	while (p != NULL) {
	    pre = p;
	    p = p->next;
	}
	pre->next = pdev;
}


/* check if the device for ip addr dst_ip already exits in device_list */
struct device_item * check_dstip(uint32 dst_ip)
{
	struct device_item * p, * pre;

	p = device_list;
	pre = NULL;

	while (p != NULL) {
	    if (p->dst_ip == dst_ip)
		return p;
	    p = p->next;
	}

	return NULL;
}


/* check if the device with name "str" already exits in device_list */
struct device_item * check_dev(char * str)
{
	struct device_item * p, * pre;

	p = device_list;
	pre = NULL;

	while (p != NULL) {
	    if (strcmp(p->dev_str, str) == 0)
		return p;
	    p = p->next;
	}

	return NULL;
}


/* get the current time */
double get_time()
{
	struct timeval tv;
	struct timezone tz;
	double cur_time; 

	if (gettimeofday(&tv, &tz) < 0) {
	    perror("get_time() fails, exit\n");
	    exit(1);
	}

	cur_time = (double)tv.tv_sec + ((double)tv.tv_usec/(double)1000000.0);
	return cur_time;
}


/* check if the packet is what we wants */
struct filter_item * in_session(struct device_item * pdev, 
		                struct TcpPacket * p)
{
	struct filter_item * ptr;
    	
	pthread_mutex_lock(&(pdev->filter_mutex));

	ptr = pdev->filter_list;;
    	while (ptr != NULL) {
            if ((p->ip.ip_src == ptr->src_ip) 
		&& (p->ip.ip_dst == ptr->dst_ip)
     		&& (p->ip.ip_p == IPPROTOCOL_TCP)
	    	&& (ntohl(p->tcp.tcp_seq) <= ptr->probe_num) 
		&& (p->tcp.tcp_flags == (TCPFLAGS_FIN))) {

	        pthread_mutex_unlock(&(pdev->filter_mutex));
		return ptr;
	    }
	    ptr = ptr->next;
	}

	pthread_mutex_unlock(&(pdev->filter_mutex));

	return NULL;
}


/* check if we have got all the packets for client's one-phase probing 	*/
/* and can send back the filter packet information 			*/
void phase_finish(struct device_item * pdev)
{
	/* Phase end condition:
	 * 1. get probe_num probing packets
	 * 2. no new packet after PhaseWait time
	 */

	double cur_time = get_time();
	struct filter_item * ptr;

	char msg_buf[64];
	int msg_len;
	int data_size; 

	pthread_mutex_lock(&(pdev->filter_mutex));

	for (ptr=pdev->filter_list; ptr != NULL; ptr=ptr->next) {
	    if (!(ptr->working)) continue;

	    if (ptr->count == 0) { 
	 	if (cur_time - ptr->pre_time > NoDataWait) {
		    /* this is a no data phase */
		    ptr->nodata_count ++;
		    if (ptr->nodata_count >= MaxNoDataNum) {
			close(ptr->control_sock);
			ptr->working = 0;
			ptr->nodata_count = 0;
			if (verbose) {
			    printf("%d no data phases, close socket with %s\n", 
				    MaxNoDataNum, ptr->src_ip_str);
			}
		    }
		    else {
			if (verbose) {
			    printf("start sending back 0 data\n");
			}

			data_size = sizeof(struct pkt_rcd_t) * ptr->count;		
			/* send the the control information to tell how many 
			 * record will be transmitted back */
			sprintf(msg_buf, "$%d$", data_size);
			msg_len = strlen(msg_buf);
#ifdef LINUX
			send(ptr->control_sock, msg_buf, msg_len, MSG_NOSIGNAL);
#else
			send(ptr->control_sock, msg_buf, msg_len, 0);
#endif
		
			if (debug) {
			    printf("send out msg: | %s |\n", msg_buf);
			}

			/* send back the record information */
#ifdef LINUX
			send(ptr->control_sock, (void *)ptr->record, 
				data_size, MSG_NOSIGNAL);
#else
			send(ptr->control_sock, (void *)ptr->record, 
				data_size, 0);
#endif

	        	ptr->pre_time = get_time();
		    }
		}
		continue;
	    }

	    /* we have got some trace packets */
	    if ((ptr->count == ptr->probe_num) || 
		(cur_time - ptr->pre_time > PhaseWait)) {

		/* clear previous bad record */
		ptr->nodata_count = 0;

		if (verbose) {
		    printf("start sending back data\n");
		}

		data_size = sizeof(struct pkt_rcd_t) * ptr->count;		

		/* send the the control information to tell how many record
		 * will be transmitted back */
		sprintf(msg_buf, "$%d$", data_size);
		msg_len = strlen(msg_buf);
#ifdef LINUX
		send(ptr->control_sock, msg_buf, msg_len, MSG_NOSIGNAL);
#else
		send(ptr->control_sock, msg_buf, msg_len, 0);
#endif
	
		if (debug) {
		    printf("send out msg: | %s |\n", msg_buf);
		}

		/* send back the record information */
#ifdef LINUX
		send(ptr->control_sock, (void *)ptr->record, 
			data_size, MSG_NOSIGNAL);
#else
		send(ptr->control_sock, (void *)ptr->record, 
			data_size, 0);
#endif

		/* ready for the next phase */
		ptr->count = 0;
	        ptr->pre_time = get_time();
	    }
	}

	pthread_mutex_unlock(&(pdev->filter_mutex));
}


/* the main function of child process, filter probing packets from clients */
void get_packets(struct device_item * pdev)
{
	pcap_t * pc;
	struct TcpPacket *pkt;
	struct PacketInfo pi;
	struct filter_item *ptr;

	if (verbose) {
	    printf("come into get_packets \n");
	}

	pc = pdev->pc;

	for (;;) {
	    /* get packets */
	    pthread_mutex_lock(&(pdev->pc_mutex));
    	    pkt =(struct TcpPacket *)CaptureGetPacket(pdev->pc, &pi);
	    pthread_mutex_unlock(&(pdev->pc_mutex));

    	    if (pkt != NULL) {
		if (debug) {
	            struct in_addr src_in, dst_in;
		    src_in.s_addr = pkt->ip.ip_src;
		    dst_in.s_addr = pkt->ip.ip_dst;
		    if (debug) {
		    	printf("get pkt: %s %s \n", 
				inet_ntoa(src_in), inet_ntoa(dst_in));
		    }
		}

		if ((ptr = in_session(pdev, pkt)) != NULL) {
		    /* it is what we want */

		    struct pkt_rcd_t * p_record = &(ptr->record[ptr->count]);

		    p_record->seq = ntohl(pkt->tcp.tcp_seq);
		    p_record->sec = pi.ts.tv_sec; 
		    p_record->u_sec = pi.ts.tv_usec;

		    if (debug) {
		     	printf("%d %d.%d  %s -> %s %d %d \n", 
			    ptr->count, 
			    p_record->sec,
			    p_record->u_sec,
			    ptr->src_ip_str,
			    ptr->dst_ip_str,
		            ntohl(pkt->tcp.tcp_seq),
			    (int)ntohs(pkt->ip.ip_len));
		    }
		    ptr->count ++;
		    ptr->pre_time = get_time();
		}
	    }

	    phase_finish(pdev);
	}
}


/*
 * If the interface for dst_ip current is not working, create a new dev_item
 * and return it; otherwise, just return the pointer to the existing dev_item
 */
struct device_item * 
update_dev_list(uint32 src_ip, uint32 dst_ip, uint32 src_port, uint32 dst_port)
{
	char * cur_dev_str;
	char errbuf[1024];
	struct device_item *pdev = NULL;

	/* device_list does not need mutex, since there is not possibility
	 * for this function to be reentry */

	/* check if the interface is already working */
	if ((pdev = check_dstip(dst_ip)) != NULL) {
	    return pdev;
	}

	cur_dev_str = pcap_lookupdev_withIP(dst_ip, errbuf);

	if (verbose) {
	    printf("request device name: %s\n", cur_dev_str);
	}

	if ((pdev = check_dev(cur_dev_str)) == NULL) {
	    if (verbose) {
	    	printf("new device %s\n", cur_dev_str);
	    }

	    /* create the a new device item */
	    pdev = (struct device_item *) malloc(sizeof(struct device_item));
	    pdev->next = NULL;
	    strcpy(pdev->dev_str, cur_dev_str);
	    pdev->dst_ip = dst_ip;

	    pdev->pc = CaptureInit_withDev(cur_dev_str, 
				src_ip, src_port, 
				dst_ip, dst_port);
	    pthread_mutex_init(&(pdev->pc_mutex), NULL);

	    if (verbose) {
		printf("device inited\n");
	    }

	    pdev->filter_list = NULL;
	    pthread_mutex_init(&(pdev->filter_mutex), NULL);

            pthread_create(&(pdev->thread), NULL, 
	    		    (void *)&get_packets, (void *)pdev);

	    append_device(pdev);
	}

	return pdev;
}


/*
 * If the <src_ip, dst_ip> pair is already in the pdev list, then just 
 * update its active_time, and return the point; 
 * otherwise, create a new filter item, and initilize it
 */
void update_filter_list(struct device_item * pdev,
		   int newsd,
		   int probe_num,
		   char * src_ip_str, 
		   char * dst_ip_str)
{
	struct filter_item * p, * pre;
	double cur_time;

	pthread_mutex_lock(&(pdev->filter_mutex));

        /* check whether any of the filter item has been idle long
         * enough to be considered "finish", or have no data to receive */
        p = pdev->filter_list;
        pre = NULL;
        cur_time = get_time();
        while (p != NULL) {
            if ((cur_time - p->pre_time > CLIENTTIMEOUT)
		|| (p->nodata_count >= MaxNoDataNum)) {
                /* delete this item */
                if (verbose) {
                    printf("delete filter item %s %s \n",
                            p->src_ip_str, p->dst_ip_str);
                }

                if (pre == NULL)
                    pdev->filter_list = pdev->filter_list->next;
                else
                    pre->next = p->next;

		if (p->working)
		    close(p->control_sock);
                free(p);

                if (pre == NULL)
                    p = pdev->filter_list;
                else
                    p = pre->next;
            }
            else {
                pre = p;
                p = p->next;
            }
        }

        /* check if src & dst ip pair already in the filter list */
        p = pdev->filter_list;
        pre = NULL;
        while (p != NULL) {
            if ((strcmp(p->src_ip_str, src_ip_str) == 0)
                && (strcmp(p->dst_ip_str, dst_ip_str) == 0)) {
                break;
            }
            else {
                pre = p;
                p = p->next;
            }
        }

        /* for new ip pair, append in the filter_list */
        if (p != NULL) {
	    if (verbose) {
		printf("filter item already exits\n");
	    }
	    if (p->working)
	        close(p->control_sock);
	    p->control_sock = newsd;
            p->pre_time = get_time();
            p->count = 0;
            p->probe_num = probe_num;
	    p->nodata_count = 0;
	    p->working = 1;
	}
	else {
	    if (verbose) {
	        printf("create a new filter item\n");
	    }
            p = (struct filter_item *)malloc(sizeof(struct filter_item));
            strcpy(p->src_ip_str, src_ip_str);
            strcpy(p->dst_ip_str, dst_ip_str);
            p->src_ip = inet_addr(src_ip_str);
            p->dst_ip = inet_addr(dst_ip_str);
	    p->control_sock = newsd;
            p->pre_time = get_time();
            p->count = 0;
            p->probe_num = probe_num;
            p->nodata_count = 0;
	    p->working = 1;
            p->next = NULL;

            if (pre == NULL) {
	    	/* this is the first filter, don't need to worry about
		 * pc init */
                pdev->filter_list = p;
	    }
            else {
                uint32 cur_src_ip[MaxClientNum];
                uint32 cur_dst_ip[MaxClientNum];
                uint32 cur_src_port[MaxClientNum];
                uint32 cur_dst_port[MaxClientNum];
                int cur_count;
                struct filter_item * cur_p;

                pre->next = p;

                /* reset the filter cmd */
                cur_p = pdev->filter_list;
                cur_count = 0;
                while (cur_p != NULL) {
                    cur_src_ip[cur_count] = cur_p->src_ip;
                    cur_dst_ip[cur_count] = cur_p->dst_ip;
                    cur_src_port[cur_count] = 0;
                    cur_dst_port[cur_count] = 0;

                    cur_count ++;
                    cur_p = cur_p->next;
                }

	        pthread_mutex_lock(&(pdev->pc_mutex));
                CaptureReset(pdev->pc,
                             cur_src_ip, cur_src_port,
                             cur_dst_ip, cur_dst_port,
                             cur_count);
	        pthread_mutex_unlock(&(pdev->pc_mutex));
            }
	}

	pthread_mutex_unlock(&(pdev->filter_mutex));
}


/* check if there is any new client making connection */
void check_client()
{	    
	    char msg_buf[1024];
	    int msg_len;
	    int newsd;

	    src_size = sizeof(struct sockaddr);
	    newsd = accept(sock, (struct sockaddr *) &src_addr, &src_size);
	    if (newsd <= 0) return;

	    /* get src addresses */
	    strcpy(src_ip_str, inet_ntoa(src_addr.sin_addr));
	    src_ip = inet_addr(src_ip_str);

	    /* get the dst ip address that client is connecting with */
	    dst_size = sizeof(struct sockaddr);
	    getsockname(newsd, (struct sockaddr *)&dst_addr, &dst_size);
	    strcpy(dst_ip_str, inet_ntoa(dst_addr.sin_addr));
	    dst_ip = inet_addr(dst_ip_str);

	    if (verbose) {
	        printf("get new connection \n");
	        printf("src_ip_str = %s \n", src_ip_str);
	        printf("dst_ip_str = %s \n", dst_ip_str);
	        printf("waiting for START msg\n");
	    }

	    /* waiting for "$START$%d$", for purpose of sanity check.
	     * Here, the %d is the probe_num set by probing client */
	    msg_len = 0;
	    while ((msg_len <= 7) || (msg_buf[msg_len-1] != '$')) {
		int tmp_len; 

		/* this is to accommodate a weird behavior of FreeBSD3.4 */
	        tmp_len = read(newsd, msg_buf + msg_len,
		                   sizeof(msg_buf)-msg_len);
	        if (tmp_len <= 0) continue;

	        msg_len += tmp_len;
	    }
	    msg_buf[msg_len] = 0;
	    
	    if (strncmp(msg_buf+1, "START", 5) == 0) {
		char num_str[16], *p;
		struct device_item *pdev = NULL;

		/* get out the probe_num */
		p = index(msg_buf + 7, '$');
		strncpy(num_str, msg_buf + 7, p - msg_buf - 7);
		num_str[p - msg_buf - 7] = 0;
		probe_num = atoi(num_str);
		if (verbose) {
		    printf("probe_num = %d \n", probe_num);
		}
		
		pdev = update_dev_list(src_ip, dst_ip, src_port, dst_port);

	 	update_filter_list(pdev, newsd, probe_num,
				src_ip_str, dst_ip_str);

#ifdef LINUX
		send(newsd, "$READY$", 7, MSG_NOSIGNAL);
#else
		send(newsd, "$READY$", 7, 0);
#endif
	    }
}


int main(int argc, char *argv[])
{
	int opt;

	RETSIGTYPE (*oldhandler)(int);

	verbose = 0;
	while ((opt = getopt(argc, argv, "dhv")) != EOF) {
	    switch (opt) {
	    case 'd':
		debug = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'h':
	    default:
		Usage();
	    }
	}

	/* Setup signal handler for cleaning up */
	(void)setsignal(SIGTERM, cleanup);
	(void)setsignal(SIGINT, cleanup);
	if ((oldhandler = setsignal(SIGHUP, cleanup)) != SIG_DFL)
	    (void)setsignal(SIGHUP, oldhandler);

	/* Create dst socket  */
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
       		perror("Unable to open socket for listener (2):"); 
		exit(1);
	}

	dst_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	dst_addr.sin_family = AF_INET;

   	dst_port = START_PORT + 1;
	dst_addr.sin_port = htons(dst_port);	
        while ((dst_port < END_PORT) && (bind(sock, 
	       (struct sockaddr *)&dst_addr, sizeof(dst_addr)) < 0)) {
   	    dst_port ++;
	    dst_addr.sin_port = htons(dst_port);	
	}

	if (verbose) {
	    printf("server port = %d \n", dst_port);
	}

	listen(sock, 5);

	/* starts listening command from clients */
	for (;;) {
	    check_client();
	}

	close(sock);
}




